// 002 u_c_points_editor
// 14 apr 2005

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_points_editor;
  interface
    uses Classes // tShiftState
        , ExtCtrls // tPanel
        , StdCtrls // tButton
        , Controls // alTop, tMouseButton
        , Buttons
        , ComCtrls // tUpDown
        , u_c_basic_object

        , u_c_2d_c_point_list
        , u_c_2d_selection_point_list

        , u_c_transformer
        , u_c_poly_line_transformer
        , u_c_bezier_transformer

        , u_c_rectangle_selecter
      ;

    type c_points_editor= class(c_basic_object)
                                m_c_container_panel_ref: tPanel;

                                m_c_control_panel: tPanel;

                                m_c_selected_label: tLabel;
                                m_c_reset_button: tButton;
                                m_c_add_speed_button: tSpeedButton;
                                m_c_delete_button: tButton;
                                m_c_bezier_checkbox: tCheckBox;
                                m_c_edit: tEdit;
                                m_c_updown: tUpDown;

                                m_c_paintbox: tPaintBox;

                                m_scale: Double;

                                m_c_2d_point_list: c_2d_selection_point_list;
                                m_c_2d_transform_point_list: c_2d_point_list;

                                m_c_poly_line_transformer: c_transformer;
                                m_c_rectangle_selecter: c_rectangle_selecter;

                                Constructor create_points_editor(p_name: String;
                                    p_c_container_panel_ref: tPanel);

                                procedure create_transform;

                                procedure handle_reset_click(p_c_sender: tObject);
                                procedure handle_speed_button_click(p_c_sender: tObject);
                                procedure handle_delete_click(p_c_sender: tObject);
                                procedure handle_bezier_checkbox_click(p_c_sender: tObject);
                                procedure handle_edit_change(Sender: TObject);

                                procedure handle_paintbox_mouse_move_event(Sender: TObject;
                                    Shift: TShiftState; X, Y: Integer);
                                procedure handle_paintbox_mouse_down_event(Sender: TObject;
                                    Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
                                procedure handle_paintbox_mouse_up_event(Sender: TObject;
                                    Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

                                procedure handle_paintbox_paint(p_c_sender: tObject);

                                function f_scaled_bezier_value(p_value: Double): Double;

                                Destructor Destroy; Override;
                              end; // c_points_editor

  implementation
    uses SysUtils, Graphics, Math
        , u_c_display, u_integer, u_double
        , u_panel
        , u_button
        , u_checkbox
        , u_label
        , u_edit, u_updown
        , u_paint_box

        ;

    // -- c_points_editor

    Constructor c_points_editor.create_points_editor(p_name: String;
         p_c_container_panel_ref: tPanel);
      var l_running_left: Integer;

      function f_left(p_width: Integer): Integer;
        begin
          Result:= l_running_left;
          Inc(l_running_left, p_width+ 5);
        end; // f_left

      begin // create_points_editor
        Inherited create_basic_object(p_name);

        m_c_container_panel_ref:= p_c_container_panel_ref;

        // -- add a control panel
        m_c_control_panel:= f_c_create_panel('',
            m_c_container_panel_ref, m_c_container_panel_ref,
            alTop,  -1, 22, clBtnFace);

        l_running_left:= 5;
        m_c_selected_label:= f_c_create_label('label', '-1',
            m_c_container_panel_ref, m_c_control_panel,
            alNone, f_left(20), 2, 16);
        m_c_selected_label.Color:= clYellow;

        m_c_reset_button:= f_c_create_button('reset', 'Reset',
            m_c_container_panel_ref, m_c_control_panel,
            f_left(40), 2, 40, 18);
        m_c_reset_button.OnClick:= handle_reset_click;

        m_c_add_speed_button:= f_c_create_speed_button('speed', 'add',
            m_c_container_panel_ref, m_c_control_panel,
            f_left(40), 2, 40, 18, True, 33);
        m_c_add_speed_button.OnClick:= handle_speed_button_click;

        m_c_delete_button:= f_c_create_button('delete', 'Delete',
            m_c_container_panel_ref, m_c_control_panel,
            f_left(40), 2, 40, 18);
        m_c_delete_button.OnClick:= handle_delete_click;

        m_c_bezier_checkbox:= f_c_create_checkbox('bezier', 'bezier',
            m_c_container_panel_ref, m_c_control_panel,
            f_left(40), 2, 50, 18);
        m_c_bezier_checkbox.OnClick:= handle_bezier_checkbox_click;

        m_c_edit:= f_c_create_edit('scale', '',
            m_c_container_panel_ref, m_c_control_panel,
            f_left(30), 2, 30, 18);
        m_c_edit.OnChange:= handle_edit_change;

        m_c_updown:= f_c_create_updown('updown',
            m_c_container_panel_ref, m_c_control_panel,
            f_left(15), 2, 15, 18,   -5, 5, 1, m_c_edit);

        // -- add the paintbox
        m_c_paintbox:= f_c_create_paint_box('paint',
            Nil, m_c_container_panel_ref,
            alClient, -1, - 1, clWhite);
        with m_c_paintbox do
        begin
          OnMouseMove:= handle_paintbox_mouse_move_event;
          OnMouseDown:= handle_paintbox_mouse_down_event;
          OnMouseUp:= handle_paintbox_mouse_up_event;
          OnPaint:=handle_paintbox_paint;
        end;

        m_c_rectangle_selecter:= c_rectangle_selecter.create_rectangle_selecter('sel', m_c_paintbox, 5);
        m_c_rectangle_selecter.m_is_active:= True;

        m_c_poly_line_transformer:= c_poly_line_transformer.create_poly_line_transformer('poly');

        // -- create_transform;
        handle_reset_click(Nil);
        m_scale:= 1.0;
      end; // create_points_editor

    procedure c_points_editor.handle_reset_click(p_c_sender: tObject);
      begin
        m_c_2d_point_list.Free;
        m_c_2d_point_list:= c_2d_selection_point_list.create_2d_selection_point_list('2d_point_list',
            m_c_paintbox.Canvas);

        m_c_add_speed_button.Down:= True;
        handle_speed_button_click(Nil);

        m_c_paintbox.Repaint;
      end; // handle_reset_click

    procedure c_points_editor.handle_speed_button_click(p_c_sender: tObject);
      begin
        with m_c_add_speed_button do
          if Down
            then begin
                Caption:= 'add';
                m_c_control_panel.Color:= clRed;
              end
            else begin
                Caption:= 'move';
                m_c_control_panel.Color:= clLime;
              end;
      end; // handle_speed_button_click

    procedure c_points_editor.handle_delete_click(p_c_sender: tObject);
      var l_index_of: Integer;
      begin
        with m_c_2d_point_list do
          if m_c_2d_selected_point_ref<> Nil
            then begin
                // -- remember another point for the current selection
                l_index_of:= f_index_of_c_2d_point(m_c_2d_selected_point_ref);

                // -- erase the point
                if m_c_2d_selected_point_ref<> Nil
                  then m_c_2d_selected_point_ref.draw_2d_selection_point(m_c_paintbox.Canvas, clWhite);
                // -- erase the transform
                if m_c_2d_transform_point_list<> Nil
                  then m_c_2d_transform_point_list.draw_2d_point_list_poly_line(m_c_paintbox.Canvas, clWhite);

                delete_2d_c_point(m_c_2d_selected_point_ref);

                if l_index_of>= f_2d_point_count
                  then Dec(l_index_of);
                if l_index_of>= 0
                  then begin
                      m_c_2d_selected_point_ref:= c_2d_selection_point(f_c_2d_point(l_index_of));
                      draw_selected_point;
                    end;
                // -- draw the transform
                create_transform;
              end;
      end; // handle_delete_click

    procedure c_points_editor.handle_bezier_checkbox_click(p_c_sender: tObject);
      begin
(*
        // -- erase the curve
        if m_c_2d_transform_point_list<> Nil
          then m_c_2d_transform_point_list.draw_2d_point_list_poly_line(m_c_paintbox.Canvas, clWhite);
*)
        // -- here adds the max hyperbole, so instead of erasing,
        // --   the best is to redraw all with Repaint

        m_c_poly_line_transformer.Free;
        if m_c_bezier_checkbox.Checked
          then m_c_poly_line_transformer:= c_bezier_transformer.create_bezier_transformer('bezier')
          else m_c_poly_line_transformer:= c_poly_line_transformer.create_poly_line_transformer('poly');

        m_c_paintbox.Repaint;

        create_transform;
      end; // handle_bezier_checkbox_click

    procedure c_points_editor.handle_edit_change(Sender: TObject);
      var l_scale: Double;
      begin
        // -- todo
        Exit;

        l_scale:= f_string_to_integer_default(m_c_edit.Text, 1);
        if l_scale<> 1
          then l_scale:= Power(2, l_scale);

        if (l_scale<> m_scale) and (m_c_paintbox<> Nil)
          then begin
              m_scale:= l_scale;
            end;
      end; // handle_edit_change

    procedure c_points_editor.create_transform;
      begin
        // -- compute the bezier function
        m_c_2d_transform_point_list.Free;

        if m_c_2d_point_list.f_2d_point_count> 1
          then begin
              m_c_2d_transform_point_list:= m_c_poly_line_transformer.f_c_transform(m_c_2d_point_list, 5);
              // display('cnt '+ IntToStr(m_c_2d_transform_point_list.f_2d_point_count));
              // m_c_2d_transform_point_list.display_2d_point_list;
              m_c_2d_transform_point_list.draw_2d_point_list_poly_line(m_c_paintbox.Canvas, clBlack);
            end
          else m_c_2d_transform_point_list:= Nil;
      end; // create_transform

    const k_radius= 3;
    var g_mouse_start_x, g_mouse_start_y: Integer;
        g_moving: Boolean= False;
        // g_pt_2d_nearest_point: t_pt_2d_point;

    procedure c_points_editor.handle_paintbox_mouse_down_event(Sender: TObject; Button: TMouseButton;
        Shift: TShiftState; X, Y: Integer);
      var l_c_new_point: c_2d_point;
      begin
        if m_c_add_speed_button.Down
          then begin
              if m_c_2d_transform_point_list<> Nil
                then m_c_2d_transform_point_list.draw_2d_point_list_poly_line(m_c_paintbox.Canvas, clWhite);

              m_c_2d_transform_point_list.Free;
              m_c_2d_transform_point_list:= Nil;

              l_c_new_point:= c_2d_selection_point.create_2d_selection_point('', X, Y, k_radius);
              with m_c_2d_point_list do
              begin
                add_and_sort_x_2d_point('', l_c_new_point);

                // -- this point becomes the current point
                // --  (mainly for deletion)
                if m_c_2d_selected_point_ref<> Nil
                  then begin
                      m_c_2d_selected_point_ref.m_is_selected:= False;
                      m_c_2d_selected_point_ref.draw_2d_selection_point(m_c_paintbox.Canvas, clBlack);
                    end;
                m_c_2d_selected_point_ref:= c_2d_selection_point(l_c_new_point);
                with c_2d_selection_point(l_c_new_point) do
                begin
                  m_is_selected:= true;
                  draw_2d_selection_point(m_c_paintbox.Canvas, clBlack);
                end;
              end; // with m_c_2d_point_list
              
              create_transform;
            end
          else begin
              g_mouse_start_x:= X; g_mouse_start_y:= Y;
              g_moving:= True;

              m_c_2d_point_list.f_c_nearest_point(g_mouse_start_x, g_mouse_start_y);

              if m_c_2d_point_list.m_c_2d_selected_point_ref<> Nil
                then begin
                    // m_c_2d_point_list.draw_selected_point;
                    with m_c_2d_point_list do
                      m_c_selected_label.Caption:= IntToStr(f_index_of_c_2d_point(m_c_2d_selected_point_ref));

                    // -- erase Bezier curve
                    if m_c_2d_transform_point_list<> Nil
                      then begin
                          m_c_2d_transform_point_list.draw_2d_point_list_poly_line(m_c_paintbox.Canvas, clWhite);
                          // -- will be freed before recreates the transformer
                        end;
                  end
                else begin
                    m_c_selected_label.Caption:= 'NO';

                    m_c_rectangle_selecter.f_handled_mouse_down(Button, Shift, X, Y)
                  end;
            end;
      end; // m_c_paintboxMouseDown

    procedure c_points_editor.handle_paintbox_mouse_move_event(Sender: TObject; Shift: TShiftState; X,
          Y: Integer);

      procedure move_points_inside_rectangle_selecter;
        var l_point_index: Integer;
        begin
          with m_c_2d_point_list do
          begin
            // -- delete rectangle_selecter (mouse droolings)
            m_c_rectangle_selecter.draw_selecter;
            // -- delete
            if m_c_2d_transform_point_list<> Nil
              then m_c_2d_transform_point_list.draw_2d_point_list_poly_line(m_c_paintbox.Canvas, clWhite);

            for l_point_index:= 0 to f_2d_point_count- 1 do
              with f_c_2d_point(l_point_index) do
                if m_c_rectangle_selecter.f_is_in_selecter(Round(m_x), Round(m_y))
                  then begin
                      // -- delete the point
                      draw_2d_point(m_c_paintbox.Canvas, clWhite, k_radius);

                      // -- move the point
                      // display('move '+ IntToStr(l_point_index));
                      m_c_rectangle_selecter.move_point(m_x, m_y);
                      // -- draw the point
                      draw_2d_point(m_c_paintbox.Canvas, clBlack, k_radius);
                    end;

            // -- draw
            create_transform;
            // -- draw rectangle_selecter
            m_c_rectangle_selecter.draw_selecter;
          end // with m_c_2d_point_list
        end; // move_points_inside_rectangle_selecter

      begin // m_c_paintboxMouseMove
        // -- only in "move" mode
        if m_c_add_speed_button.Down
          then Exit;

        if (ssLeft in Shift) and (m_c_2d_point_list.m_c_2d_selected_point_ref<> Nil)
          then
            with m_c_2d_point_list.m_c_2d_selected_point_ref do
            begin
              // display('move '+ f_display_2d_point);
              // -- delete
              draw_2d_selection_point(m_c_paintbox.Canvas, clWhite);

              // -- move
              m_x:= X;
              m_y:= Y;

              // -- draw
              draw_2d_selection_point(m_c_paintbox.Canvas, clBlack);
            end // with m_c_2d_point_list
          else
            if m_c_rectangle_selecter.f_handled_mouse_move(Shift, X, Y)
              then begin
                  if ssRight in Shift
                    then move_points_inside_rectangle_selecter;
                  Exit;
                end;
      end; // m_c_paintboxMouseMove

    procedure c_points_editor.handle_paintbox_mouse_up_event(Sender: TObject; Button: TMouseButton;
          Shift: TShiftState; X, Y: Integer);
      begin
        if m_c_add_speed_button.Down
          then Exit;

        if g_moving
          then begin
              create_transform;
              g_moving:= False;
            end;
      end; // m_c_paintboxMouseUp

    procedure c_points_editor.handle_paintbox_paint(p_c_sender: tObject);
      var l_x_fat, l_y_fat: Integer;
          l_width, l_height: Integer;

      procedure draw_hyperbole;
        var l_index: Integer;

        function f_x: Integer;
          begin
            Result:= l_x_fat+ (l_width- 2* l_x_fat)* l_index div 10;
          end; // f_x

        function f_y: Integer;
          var l_1_over_x: Double;
          begin
            l_1_over_x:= 1/ (l_index / 10);
            // display('  y '+ f_display_double(l_1_over_x- 1, 3, 3));

            Result:= l_height div 4+ Round((l_height div 2)* (1- l_1_over_x/ 2));
          end; // f_y

        begin // draw_hyperbole
          with m_c_paintbox, Canvas do
          begin
            l_index:= 10;
            MoveTo(f_x, f_y);
            for l_index:= 9 downto 2 do
            begin
              LineTo(f_x, f_y);
              // display(IntToStr(f_x)+ ' '+ IntToStr(f_y));
            end;
          end;
        end; // draw_hyperbole

      begin // handle_paintbox_paint
        with m_c_paintbox, Canvas do
        begin
          if m_c_poly_line_transformer is c_bezier_transformer
            then begin
                l_x_fat:= Width div 10;
                l_width:= Width; l_height:= Height;

                l_y_fat:= Height div 4;

                Rectangle(l_x_fat, l_y_fat, Width- l_x_fat, Height- l_y_fat);
                MoveTo(0, Height div 2); LineTo(Width, Height div 2);
                
                draw_hyperbole;
              end
            else begin
                MoveTo(0, Height div 2); LineTo(Width, Height div 2);
              end;

          m_c_2d_point_list.draw_2d_point_list_points(Canvas, clBlack, k_radius);
        end;

        if m_c_rectangle_selecter.m_is_active
          then m_c_rectangle_selecter.draw_selecter;
      end; // handle_paintbox_paint

    function c_points_editor.f_scaled_bezier_value(p_value: Double): Double;
        // -- p_value between 0 and 1
      var l_x_value, l_y_value: Double;
      begin
        // -- find the x value
        with m_c_paintbox do
        begin
          l_x_value:= Width div 10+ Width * 0.8* p_value;
          l_y_value:= m_c_poly_line_transformer.f_interpolate_value(m_c_2d_point_list, l_x_value);
          // -- scale back, with a 1/4 factor
          Result:= 1+ (Height/ 2- l_y_value)/ (Height/ 4);
        end;
      end; // f_scaled_bezier_value

    Destructor c_points_editor.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_points_editor
    end. // u_c_points_editor
